//---------------------------------------------------------------------------

#include "stdafx.h"

#ifdef _MSC_VER
#pragma hdrstop
#endif

using namespace std;

#include "../../Kernel/kernelincludes.h"

#include "defaults.h"
// #include "TimeStamp.h"
#include "CovMatrix.h"
#include "HeurCover.h"
#include "AuraCover.h"
#include "LagBaBCover.h"
#include "Expand.h"

#include "Cover.h"
//---------------------------------------------------------------------------

Cover::Cover( PLA *_source, PLA *_implpool )
{
    source = _source;
    implpool = _implpool;
}

void Cover::GenerateOM(void)
// Generates OM values in the implicant pool (only from void "~" values)
{
    unsigned int i;
    TermList::iterator ti;

    for ( ti = implpool->getData()->begin(); ti != implpool->getData()->end(); ti++ )
        for ( i = 0; i < implpool->outputs(); i++ )
			if ( (*ti).isOM_undef(i) ) {
                if ( source->isImplicant( (*ti), i ) ) (*ti).setOM_1(i);
                else (*ti).setOM_0(i);
			}
}

void Cover::GenerateCoverVector( Term &t )
/*
	Generates CoverVector structure for a given term (if it is empty)
	It's size is equal to the # of 1's in the source OM (taken from PLA::getOnSet)
*/
{
	vector<bitvector> *SO_covvct;
	CubeList::iterator ti;
	unsigned int o;
	bool covered;		// is the term covered (contained) by t?
	bitvector *covvct;
	
	if ( !t.getCoverVector()->empty() && !t.getSO_CoverVector(0)->empty() ) return;

	SO_covvct = t.getSO_CoverVector();
	covvct = t.getCoverVector();
	SO_covvct->resize( source->outputs() );
	for ( o = 0; o < source->outputs(); o++ )
		for ( ti = source->getOnSet(o)->getData()->begin(); ti != source->getOnSet(o)->getData()->end(); ti++ ) {
			covered = t.Cube::contains(*ti);
			if ( covered && t.isOM_1(o) ) {		// covers this output and it is an implicant of o-th function
				covvct->push_back(true);
				(*SO_covvct)[o].push_back(true);
			} else {
				covvct->push_back(false);
				(*SO_covvct)[o].push_back(false);
			}
		}
}

void Cover::GenerateCoverVectors(void)
/*
	Generates CoverVectors for the whole implicant pool
*/
{
	TermList::iterator ti;

	for ( ti = implpool->getData()->begin(); ti != implpool->getData()->end(); ti++ )
		GenerateCoverVector(*ti);
}

void Cover::RemoveAbsorbedByCover(void)
/*
	By comparing CoverVectors, remove clearly absorbed terms from the pool.
	If same terms are covered by two, the one with smaller dimension is removed.
*/
{
	TermList::iterator ti1, ti2;
	bool abs1, abs2;				// which one is absorbed
	bool deleted;

	ti1 = implpool->getData()->begin();
	do {
		if ( ti1->getCoverVector()->empty() ) GenerateCoverVector(*ti1);
		deleted = false;
		ti2 = ti1;
		ti2++;
		while ( ti2 != implpool->getData()->end() ) {
			if ( ti2->getCoverVector()->empty() ) GenerateCoverVector(*ti2);
			abs1 = true;
			abs2 = true;
			if ( !(*ti1).getCoverVector()->contains( *(*ti2).getCoverVector() )) abs2 = false;
			if ( !(*ti2).getCoverVector()->contains( *(*ti1).getCoverVector() )) abs1 = false;
			if ( abs1 && abs2 ) {
				if ( (*ti1).getDim() >= (*ti1).getDim() ) ti2 = implpool->DeleteTerm(ti2);
				else {
					ti1 = implpool->DeleteTerm(ti1);
					deleted = true;
					break;
				}
			}
			else if ( abs1 && !abs2 ) { 
				ti1 = implpool->DeleteTerm(ti1);
				deleted = true;
				break;
			}
			else if ( !abs1 && abs2 ) ti2 = implpool->DeleteTerm(ti2);
			else ti2++;
		}
		if ( !deleted ) ti1++;
	} while ( ti1 != implpool->getData()->end() );
}

PLA *Cover::Run(void)
// Runs the CP
{
	CovMatrix *CM;
	PLA *solution;
	unsigned int i;
	TermList::iterator ti;
	unsigned int rownum;
	list<unsigned int> *sol;
	list<unsigned int>::iterator sol_it;
	float t1;
	ofstream trialfile;

//	TS.start(TS_CP);

	ti = implpool->getData()->begin();		// determine vector size by the first one
	if ( (*ti).getCoverVector()->empty() ) GenerateCoverVectors();
	rownum = (*ti).getCoverVector()->size();
	CM = new CovMatrix( rownum, implpool->getData()->size() );
	if ( CM == NULL) throw RuntimeException( "Cover: Not enough memory" );
	for ( ti = implpool->getData()->begin(); ti != implpool->getData()->end(); ti++ ) {
		if ( (*ti).getCoverVector()->size() != rownum ) throw RuntimeException( "Cover::Run:: Bad size of cover vector" );
		for ( i = 0; i < rownum; i++ )
			CM->SetValue(i, implpool->getIndex(ti), (*(*ti).getCoverVector())[i]);
		CM->SetCost(implpool->getIndex(ti), implpool->inputs() - (*ti).getDim() + 1);
	}
	
//	CM->DumpMatrix( "matrix.dat" );

	trialfile.open("trials.csv", ios::app);
	t1 = GetMyCPUTime();

	sol = Solve(CM);

	trialfile << GetMyCPUTime()-t1 << "; ";
	trialfile.close();

	solution = new PLA(implpool);
	if ( solution == NULL) throw RuntimeException( "Cover: Not enough memory" );
	for ( sol_it = sol->begin(); sol_it != sol->end(); sol_it++ )
		solution->AppendTerm( *implpool->getTermByIndex(*sol_it) );

	delete CM;
	delete sol;			// deallocates solution allocated by CP solver

//	TS.stop(TS_CP);
	return solution;
}

void Cover::Sparse_output( PLA *pla )
// Removes unnecessary 1's from the output matrix
// (by solving m separate covering problems)
// Input of sparsed terms is expanded
// Does not modify the number of terms
{
	CovMatrix *CM;
	unsigned int i, j, k;
	TermList::iterator ti;
	unsigned int rownum, colnum;
	list<unsigned int> *sol;
	list<unsigned int>::iterator sol_it;
	vector<bool> solvector;

//	TS.start(TS_Sparse);
	for ( ti = pla->getData()->begin(); ti != pla->getData()->end(); ti++ )
		(*ti).tmp = 0;			// candidates for input expansion

	for ( i = 0; i < pla->outputs(); i++ ) {	// generate cover matrix for this output		
			// determine the number of rows by the first one
		ti = implpool->getData()->begin();		// determine vector size by the first one
		if ( (*ti).getSO_CoverVector(i)->empty() ) GenerateCoverVectors();
		rownum = (*ti).getSO_CoverVector(i)->size();
			// count the number of 1's in the OM[i]
		colnum = 0;
		for ( ti = pla->getData()->begin(); ti != pla->getData()->end(); ti++ )
			if ( (*ti).isOM_1(i) ) colnum++;
		CM = new CovMatrix( rownum, colnum );
		if ( CM == NULL) throw RuntimeException( "Cover: Not enough memory" );
		k = 0;
		for ( ti = pla->getData()->begin(); ti != pla->getData()->end(); ti++ ) {
			if ( (*ti).getSO_CoverVector(i)->size() != rownum )
				throw RuntimeException( "Cover::Sparse_output:: Bad size of cover vector" );
			if ( (*ti).isOM_1(i) ) {
				for ( j = 0; j < rownum; j++ )
					CM->SetValue( j, k, (*(*ti).getSO_CoverVector(i))[j] );
				k++;
			}
		}
			// solve it
		sol = Solve(CM);

			// substitute unnecessary 1's by ~
		solvector.clear();
		solvector.resize( colnum, false );
		for ( sol_it = sol->begin(); sol_it != sol->end(); sol_it++ )
			solvector[*sol_it] = true;
		k = 0;
		for ( ti = pla->getData()->begin(); ti != pla->getData()->end(); ti++ )
			if ( (*ti).isOM_1(i) )
				if ( !solvector[k++] ) {
					(*ti).setOM_undef(i);
					(*ti).tmp = 1;
				}
		delete CM;
		delete sol;			// deallocates solution allocated by CP solver

	}
			// expand input for sparsed terms
	for ( ti = pla->getData()->begin(); ti != pla->getData()->end(); ti++ )
		if ( (*ti).tmp == 1 ) Expand(source).ExpandTerm(*ti);

//	TS.stop(TS_Sparse);
}

list<unsigned int> *Cover::Solve( CovMatrix *CM )
{
	switch ( method ) {
		case LCMC: return HeurCover(CM).Run();
		case Aura: return AuraCover(CM).Run();
		case BBLag: return LagBaBCover(CM).Run();
		default: throw RuntimeException("Cover::Method not implemented yet");
	}
}

